import unittest
import os
import shutil
import filecmp
from ariba import cluster, common, reference_data

modules_dir = os.path.dirname(os.path.abspath(cluster.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')
cluster.unittest = True


def clean_cluster_dir(d, exclude=None):
    if not os.path.exists(d):
        return

    '''Cleans up all files made except original ones in a cluster directory'''
    keep = set(['genes.fa', 'reads_1.fq', 'reads_2.fq'])
    if exclude is not None:
        for f in exclude:
            keep.add(f)

    for name in os.listdir(d):
        if name not in keep:
            full_path = os.path.join(d, name)
            if os.path.isdir(full_path):
                common.rmtree(full_path)
            else:
                os.unlink(full_path)


class TestCluster(unittest.TestCase):
    def test_init_fail_files_missing(self):
        '''test init_fail_files_missing'''
        refdata_fa = os.path.join(data_dir, 'cluster_test_init_refdata.fa')
        meatadata_tsv = os.path.join(data_dir, 'cluster_test_init_refdata.tsv')
        refdata = reference_data.ReferenceData([refdata_fa], [meatadata_tsv])

        dirs = [
            'cluster_test_init_no_refs_fa',
            'cluster_test_init_no_reads_1',
            'cluster_test_init_no_reads_2',
        ]
        dirs = [os.path.join(data_dir, d) for d in dirs]
        for d in dirs:
            tmpdir = 'tmp.cluster_test_init_fail_files_missing'
            common.rmtree(tmpdir)
            shutil.copytree(d, tmpdir)
            with self.assertRaises(cluster.Error):
                cluster.Cluster(tmpdir, 'name', refdata=refdata, total_reads=42, total_reads_bases=4242)
            common.rmtree(tmpdir)


    def test_number_of_reads_for_assembly(self):
        '''Test _number_of_reads_for_assembly'''
        tests = [
            (50, 1000, 10, 20, 40),
            (50, 999, 10, 20, 42),
            (50, 1000, 10, 10, 20),
            (50, 1000, 10, 5, 10),
        ]

        for insert, bases, reads, coverage, expected in tests:
            self.assertEqual(expected, cluster.Cluster._number_of_reads_for_assembly(100, insert, bases, reads, coverage))


    def test_make_reads_for_assembly_proper_sample(self):
        '''Test _make_reads_for_assembly when sampling from reads'''
        reads_in1 = os.path.join(data_dir, 'cluster_test_make_reads_for_assembly.in1.fq')
        reads_in2 = os.path.join(data_dir, 'cluster_test_make_reads_for_assembly.in2.fq')
        expected_out1 = os.path.join(data_dir, 'cluster_test_make_reads_for_assembly.out1.fq')
        expected_out2 = os.path.join(data_dir, 'cluster_test_make_reads_for_assembly.out2.fq')
        reads_out1 = 'tmp.test_make_reads_for_assembly.reads.out1.fq'
        reads_out2 = 'tmp.test_make_reads_for_assembly.reads.out2.fq'
        reads_written = cluster.Cluster._make_reads_for_assembly(10, 20, reads_in1, reads_in2, reads_out1, reads_out2, random_seed=42)
        self.assertEqual(14, reads_written)
        self.assertTrue(filecmp.cmp(expected_out1, reads_out1, shallow=False))
        self.assertTrue(filecmp.cmp(expected_out2, reads_out2, shallow=False))
        os.unlink(reads_out1)
        os.unlink(reads_out2)


    def test_make_reads_for_assembly_symlinks(self):
        '''Test _make_reads_for_assembly when just makes symlinks'''
        reads_in1 = os.path.join(data_dir, 'cluster_test_make_reads_for_assembly.in1.fq')
        reads_in2 = os.path.join(data_dir, 'cluster_test_make_reads_for_assembly.in2.fq')
        reads_out1 = 'tmp.test_make_reads_for_assembly.reads.out1.fq'
        reads_out2 = 'tmp.test_make_reads_for_assembly.reads.out2.fq'
        reads_written = cluster.Cluster._make_reads_for_assembly(20, 20, reads_in1, reads_in2, reads_out1, reads_out2, random_seed=42)
        self.assertEqual(20, reads_written)
        self.assertTrue(os.path.islink(reads_out1))
        self.assertTrue(os.path.islink(reads_out2))
        self.assertEqual(os.readlink(reads_out1), reads_in1)
        self.assertEqual(os.readlink(reads_out2), reads_in2)
        os.unlink(reads_out1)
        os.unlink(reads_out2)


    def test_full_run_no_reads_after_filtering(self):
        '''test complete run of cluster when filtering removes all reads'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_no_reads_after_filtering.in.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_no_reads_after_filtering.in.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.test_full_run_no_reads_after_filtering'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_no_reads_after_filtering'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=0, total_reads_bases=0)
        c.run()

        expected = '\t'.join(['.', '.', '.', '.', '64', '0', 'cluster_name'] + ['.'] * 24)
        self.assertEqual([expected], c.report_lines)
        self.assertFalse(c.status_flag.has('ref_seq_choose_fail'))
        self.assertTrue(c.status_flag.has('assembly_fail'))
        common.rmtree(tmpdir)


    def test_full_run_choose_ref_fail(self):
        '''test complete run of cluster when choosing ref seq fails'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_choose_ref_fail.in.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_choose_ref_fail.in.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.test_full_run_choose_ref_fail'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_choose_ref_fail'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=2, total_reads_bases=108)
        c.run()

        expected = '\t'.join(['.', '.', '.', '.', '1024', '2', 'cluster_name'] + ['.'] * 24)
        self.assertEqual([expected], c.report_lines)
        self.assertTrue(c.status_flag.has('ref_seq_choose_fail'))
        self.assertFalse(c.status_flag.has('assembly_fail'))
        common.rmtree(tmpdir)


    def test_full_run_ref_not_in_cluster(self):
        '''test complete run of cluster when nearest ref is outside cluster'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_ref_not_in_cluster.in.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_ref_not_in_cluster.in.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.test_full_run_ref_not_in_cluster'
        all_refs_fa =  os.path.join(data_dir, 'cluster_test_full_run_ref_not_in_cluster.all_refs.fa')
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_ref_not_in_cluster'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=72, total_reads_bases=3600, all_ref_seqs_fasta=all_refs_fa)
        c.run()

        expected = '\t'.join(['.', '.', '.', '.', '1024', '72', 'cluster_name'] + ['.'] * 24)
        self.assertEqual([expected], c.report_lines)
        self.assertTrue(c.status_flag.has('ref_seq_choose_fail'))
        self.assertFalse(c.status_flag.has('assembly_fail'))
        common.rmtree(tmpdir)


    def test_full_run_assembly_fail(self):
        '''test complete run of cluster when assembly fails'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_assembly_fail.in.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_assembly_fail.in.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.test_full_run_assembly_fail'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_assembly_fail'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=4, total_reads_bases=304)
        c.run()

        expected = '\t'.join(['.', '.', '.', '.', '64', '4', 'cluster_name'] + ['.'] * 24)
        self.assertEqual([expected], c.report_lines)
        self.assertFalse(c.status_flag.has('ref_seq_choose_fail'))
        self.assertTrue(c.status_flag.has('assembly_fail'))
        common.rmtree(tmpdir)


    def test_full_run_ok_non_coding(self):
        '''test complete run of cluster on a noncoding sequence'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_ok_non_coding.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_ok_non_coding.metadata.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.test_full_run_ok_non_coding'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_ok_non_coding'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=72, total_reads_bases=3600)
        c.run()

        self.maxDiff=None
        expected = [
            'noncoding1\tnoncoding1\t0\t0\t531\t72\tcluster_name\t120\t120\t95.87\tcluster_name.l15.c30.ctg.1\t234\t15.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t14\t14\tA\t74\t74\tT\t19\tT\t19\tnoncoding1:0:0:A14T:.:ref has wild type, reads has variant so should report\tgeneric description of noncoding1',
            'noncoding1\tnoncoding1\t0\t0\t531\t72\tcluster_name\t120\t120\t95.87\tcluster_name.l15.c30.ctg.1\t234\t15.4\t0\t.\tn\t.\t0\tG61T\tSNP\t61\t61\tG\t121\t121\tT\t24\tT\t24\t.\tgeneric description of noncoding1',
            'noncoding1\tnoncoding1\t0\t0\t531\t72\tcluster_name\t120\t120\t95.87\tcluster_name.l15.c30.ctg.1\t234\t15.4\t0\t.\tn\t.\t0\t.82C\tINS\t82\t82\tA\t143\t143\tC\t23\tC\t23\t.\tgeneric description of noncoding1',
            'noncoding1\tnoncoding1\t0\t0\t531\t72\tcluster_name\t120\t120\t95.87\tcluster_name.l15.c30.ctg.1\t234\t15.4\t0\t.\tn\t.\t0\tT108.\tDEL\t108\t108\tT\t168\t168\tC\t17\tC\t17\t.\tgeneric description of noncoding1',
            'noncoding1\tnoncoding1\t0\t0\t531\t72\tcluster_name\t120\t120\t95.87\tcluster_name.l15.c30.ctg.1\t234\t15.4\t1\tSNP\tn\tA6G\t1\t.\t.\t6\t6\tG\t66\t66\tG\t19\tG\t19\tnoncoding1:0:0:A6G:.:variant in ref and reads so should report\tgeneric description of noncoding1',
            'noncoding1\tnoncoding1\t0\t0\t531\t72\tcluster_name\t120\t120\t95.87\tcluster_name.l15.c30.ctg.1\t234\t15.4\t1\tSNP\tn\tG9T\t0\t.\t.\t9\t9\tG\t69\t69\tG\t19\tG\t19\tnoncoding1:0:0:G9T:.:wild type in ref and reads\tgeneric description of noncoding1'
        ]

        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_ok_presence_absence(self):
        '''test complete run of cluster on a presence absence gene'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_ok_presence_absence.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_ok_presence_absence.metadata.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_presence_absence'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_ok_presence_absence'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=64, total_reads_bases=3200)
        c.run()

        expected = [
            'presence_absence1\tpresence_absence1\t1\t0\t539\t64\tcluster_name\t96\t96\t97.92\tcluster_name.l15.c30.ctg.1\t213\t15.0\t1\tSNP\tp\tA10V\t1\tA10V\tNONSYN\t28\t30\tGCG\t83\t85\tGTG\t22;22;21\tG;T;G\t22;22;21\tpresence_absence1:1:0:A10V:.:Ref has wild, reads have variant so report\tGeneric description of presence_absence1',
            'presence_absence1\tpresence_absence1\t1\t0\t539\t64\tcluster_name\t96\t96\t97.92\tcluster_name.l15.c30.ctg.1\t213\t15.0\t0\t.\tp\t.\t0\t.\tSYN\t52\t54\tATT\t107\t109\tATC\t31;31;32\tA;T;C\t31;31;32\t.\tGeneric description of presence_absence1',
            'presence_absence1\tpresence_absence1\t1\t0\t539\t64\tcluster_name\t96\t96\t97.92\tcluster_name.l15.c30.ctg.1\t213\t15.0\t1\tSNP\tp\tR3S\t0\t.\t.\t7\t9\tCGC\t62\t64\tCGC\t18;17;17\tC;G;C\t18;17;17\tpresence_absence1:1:0:R3S:.:Ref and assembly have wild type\tGeneric description of presence_absence1',
            'presence_absence1\tpresence_absence1\t1\t0\t539\t64\tcluster_name\t96\t96\t97.92\tcluster_name.l15.c30.ctg.1\t213\t15.0\t1\tSNP\tp\tI5A\t1\t.\t.\t13\t15\tGCG\t68\t70\tGCG\t18;20;20\tG;C;G\t18;20;20\tpresence_absence1:1:0:I5A:.:Ref and reads have variant so report\tGeneric description of presence_absence1',
        ]

        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_ok_variants_only_variant_not_present(self):
        '''test complete run of cluster on a variants only gene when variant not present'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_ok_variants_only.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_ok_variants_only.not_present.metadata.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_variants_only.not_present'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_ok_variants_only'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=66, total_reads_bases=3300)
        c.run()
        expected = [
            'variants_only1\tvariants_only1\t1\t1\t27\t66\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t215\t15.3\t1\tSNP\tp\tR3S\t0\t.\t.\t7\t9\tCGC\t65\t67\tCGC\t18;18;19\tC;G;C\t18;18;19\tvariants_only1:1:1:R3S:.:Ref and assembly have wild type, so do not report\tGeneric description of variants_only1'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_ok_variants_only_variant_not_present_always_report(self):
        '''test complete run of cluster on a variants only gene when variant not present but always report variant'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_ok_variants_only.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_varonly.not_present.always_report.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_full_run_varonly.not_present.always_report'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_ok_variants_only'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=66, total_reads_bases=3300)
        c.run()
        expected = [
            'variants_only1\tvariants_only1\t1\t1\t27\t66\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t215\t15.3\t1\tSNP\tp\tR3S\t0\t.\t.\t7\t9\tCGC\t65\t67\tCGC\t18;18;19\tC;G;C\t18;18;19\tvariants_only1:1:1:R3S:.:Ref and assembly have wild type, but always report anyway\tGeneric description of variants_only1'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_ok_variants_only_variant_is_present(self):
        '''test complete run of cluster on a variants only gene when variant is present'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_ok_variants_only.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_ok_variants_only.present.metadata.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_variants_only.present'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_ok_variants_only'), tmpdir)

        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=66, total_reads_bases=3300)
        c.run()

        expected = [
            'variants_only1\tvariants_only1\t1\t1\t27\t66\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t215\t15.3\t1\tSNP\tp\tR3S\t0\t.\t.\t7\t9\tCGC\t65\t67\tCGC\t18;18;19\tC;G;C\t18;18;19\tvariants_only1:1:1:R3S:.:Ref and assembly have wild type\tGeneric description of variants_only1',
            'variants_only1\tvariants_only1\t1\t1\t27\t66\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t215\t15.3\t1\tSNP\tp\tI5A\t1\t.\t.\t13\t15\tGCG\t71\t73\tGCG\t17;17;17\tG;C;G\t17;17;17\tvariants_only1:1:1:I5A:.:Ref and reads have variant so report\tGeneric description of variants_only1',
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_ok_gene_start_mismatch(self):
        '''test complete run where gene extended because too different at end for full nucmer match'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_ok_gene_start_mismatch.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_ok_gene_start_mismatch.metadata.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_gene_start_mismatch'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_ok_gene_start_mismatch'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=112, total_reads_bases=1080)
        c.run()
        expected = [
            'gene\tgene\t1\t0\t27\t112\tcluster_name\t96\t96\t100.0\tcluster_name.l6.c30.ctg.1\t362\t27.8\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\tGeneric description of gene'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_smtls_snp_presabs_gene(self):
        '''test complete run where samtools calls a snp in a presence/absence gene'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_presabs_gene.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_presabs_gene.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_samtools_snp_pres_abs_gene'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_smtls_snp_presabs_gene'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()
        expected = [
            'ref_gene\tref_gene\t1\t0\t155\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t0\tHET\t.\t.\t.\tG18A\t.\t18\t18\tG\t137\t137\tG\t63\tG,A\t32,31\t.\tGeneric description of ref_gene'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_smtls_snp_varonly_gene_2(self):
        '''test complete run where samtools calls a snp in a variant only gene'''
        # _2 because I think test_full_run_smtls_snp_varonly_gene tests the asame functionality.
        # ... but let's leave both tests in anyway
        fasta_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene_2.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene_2.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_full_run_smtls_snp_varonly_gene_2'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene_2'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()
        expected = [
            'ref_gene\tref_gene\t1\t1\t155\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t0\tHET\t.\t.\t.\tG18A\t.\t18\t18\tG\t137\t137\tG\t63\tG,A\t32,31\t.\tGeneric description of ref_gene'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_known_smtls_snp_presabs_gene(self):
        '''test complete run where samtools calls a snp at a known snp location in a presence/absence gene'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_known_smtls_snp_presabs_gene.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_known_smtls_snp_presabs_gene.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_samtools_snp_known_position_pres_abs_gene'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_known_smtls_snp_presabs_gene'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()

        # We shouldn't get an extra 'HET' line because we already know about the snp, so
        # included in the report of the known snp
        expected = [
            'ref_gene\tref_gene\t1\t0\t155\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t1\tSNP\tp\tM6I\t0\t.\t.\t16\t18\tATG\t135\t137\tATG\t65;64;63\tA;T;G,A\t65;64;32,31\tref_gene:1:0:M6I:.:Description of M6I snp\t.'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_smtls_snp_varonly_gene_no_snp(self):
        '''test complete run where samtools calls a snp at a known snp location in a variant only gene, gene does not have variant'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene_no_snp.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene_no_snp.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_smtls_snp_varonly_gene_no_snp'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene_no_snp'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()

        # We shouldn't get an extra 'HET' line because we already know about the snp, so
        # included in the report of the known snp
        expected = [
            'ref_gene\tref_gene\t1\t1\t155\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t1\tSNP\tp\tM6I\t0\t.\t.\t16\t18\tATG\t135\t137\tATG\t65;64;63\tA;T;G,A\t65;64;32,31\tref_gene:1:1:M6I:.:Description of M6I snp\t.'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_smtls_snp_varonly_gene(self):
        '''test complete run where samtools calls a snp at a known snp location in a variant only gene, gene does have variant'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_samtools_snp_known_position_var_only_gene_does_have_var'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_gene'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()

        # We shouldn't get an extra 'HET' line because we already know about the snp, so
        # included in the report of the known snp
        expected = [
            'ref_gene\tref_gene\t1\t1\t155\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t1\tSNP\tp\tI6M\t1\t.\t.\t16\t18\tATG\t135\t137\tATG\t65;64;63\tA;T;G,A\t65;64;32,31\tref_gene:1:1:I6M:.:Description of I6M snp\t.'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_smtls_snp_presabs_nonc(self):
        '''test complete run where samtools calls a snp in a presence/absence noncoding sequence'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_presabs_nonc.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_presabs_nonc.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_smtls_snp_presabs_nonc'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_smtls_snp_presabs_nonc'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()
        expected = [
            'ref_seq\tref_seq\t0\t0\t147\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t0\tHET\t.\t.\t.\tG18A\t.\t18\t18\tG\t137\t137\tG\t63\tG,A\t32,31\t.\tGeneric description of ref_seq'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_smtls_known_snp_presabs_nonc(self):
        '''test complete run where samtools calls a snp in a presence/absence noncoding sequence at a known snp position'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_smtls_known_snp_presabs_nonc.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_smtls_known_snp_presabs_nonc.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_smtls_known_snp_presabs_nonc'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_smtls_known_snp_presabs_nonc'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()
        expected = [
            'ref_seq\tref_seq\t0\t0\t147\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t1\tSNP\tn\tG18A\t0\t.\t.\t18\t18\tG\t137\t137\tG\t63\tG,A\t32,31\tref_seq:0:0:G18A:.:Description of G18A\tGeneric description of ref_seq'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_smtls_snp_varonly_nonc(self):
        '''test complete run where samtools calls a snp in a presence/absence noncoding sequence'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_nonc.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_nonc.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_full_run_smtls_snp_varonly_nonc'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_nonc'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()
        expected = [
            'ref_seq\tref_seq\t0\t1\t147\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t0\tHET\t.\t.\t.\tG18A\t.\t18\t18\tG\t137\t137\tG\t63\tG,A\t32,31\t.\tGeneric description of ref_seq'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_known_smtls_snp_presabs_nonc(self):
        '''test complete run where samtools calls a snp at a known snp location in a presence/absence noncoding'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_known_smtls_snp_presabs_nonc.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_known_smtls_snp_presabs_nonc.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_samtools_snp_known_position_pres_abs_noncoding'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_known_smtls_snp_presabs_nonc'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()

        # We shouldn't get an extra 'HET' line because we already know about the snp, so
        # included in the report of the known snp
        expected = [
            'ref_seq\tref_seq\t0\t0\t147\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t1\tSNP\tn\tG18A\t0\t.\t.\t18\t18\tG\t137\t137\tG\t63\tG,A\t32,31\tref_seq:0:0:G18A:.:Description of G18A snp\t.'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_smtls_snp_varonly_nonc_no_snp(self):
        '''test complete run where samtools calls a snp at a known snp location in a presence/absence noncoding and sample does not have the var'''
        fasta_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_nonc_no_snp.fa')
        tsv_in = os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_nonc_no_snp.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_samtools_snp_known_position_var_only_noncoding'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_full_run_smtls_snp_varonly_nonc_no_snp'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()

        # We shouldn't get an extra 'HET' line because we already know about the snp, so
        # included in the report of the known snp
        expected = [
            'ref_seq\tref_seq\t0\t1\t147\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t1\tSNP\tn\tG18A\t0\t.\t.\t18\t18\tG\t137\t137\tG\t63\tG,A\t32,31\tref_seq:0:1:G18A:.:Description of G18A snp\t.'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_cluster_test_full_run_smtls_snp_varonly_nonc(self):
        '''test complete run where samtools calls a snp at a known snp location in a presence/absence noncoding and sample has the var'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_smtls_snp_varonly_nonc.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_smtls_snp_varonly_nonc.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_ok_samtools_snp_known_position_var_only_noncoding'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_smtls_snp_varonly_nonc'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=148, total_reads_bases=13320)
        c.run()

        # We shouldn't get an extra 'HET' line because we already know about the snp, so
        # included in the report of the known snp
        expected = [
            'ref_seq\tref_seq\t0\t1\t147\t148\tcluster_name\t96\t96\t100.0\tcluster_name.l15.c30.ctg.1\t335\t39.8\t1\tSNP\tn\tA18G\t1\t.\t.\t18\t18\tG\t137\t137\tG\t63\tG,A\t32,31\tref_seq:0:1:A18G:.:Description of A18G snp\t.'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_partial_assembly(self):
        '''Test complete run where only part of the ref gene is present in the reads'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_partial_asmbly.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_partial_asmbly.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_partial_assembly'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_partial_asmbly'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=278, total_reads_bases=15020)
        c.run()

        expected = [
            'presence_absence1\tpresence_absence1\t1\t0\t19\t278\tcluster_name\t96\t77\t100.0\tcluster_name.l15.c17.ctg.1\t949\t20.5\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\tGeneric description of presence_absence1'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_multiple_vars_in_codon(self):
        '''Test complete run where there is a codon with a SNP and an indel'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_multiple_vars.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_multiple_vars.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_run_multiple_vars'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_multiple_vars'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=292, total_reads_bases=20900)
        c.run()

        expected = [
            'presence_absence1\tpresence_absence1\t1\t0\t539\t292\tcluster_name\t96\t96\t96.91\tcluster_name.l15.c30.ctg.1\t1074\t20.4\t0\t.\tp\t.\t0\t.\tMULTIPLE\t25\t26\tGA\t487\t489\tCAT\t27;26;25\tC;A;T\t27;26;25\t.\tGeneric description of presence_absence1',
            'presence_absence1\tpresence_absence1\t1\t0\t539\t292\tcluster_name\t96\t96\t96.91\tcluster_name.l15.c30.ctg.1\t1074\t20.4\t0\t.\tp\t.\t0\tA10fs\tFSHIFT\t28\t28\tG\t491\t491\tG\t26\tG\t26\t.\tGeneric description of presence_absence1',
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_delete_codon(self):
        '''Test complete run where there is a deleted codon'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_delete_codon.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_delete_codon.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_delete_codon'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_delete_codon'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=292, total_reads_bases=20900)
        c.run()

        expected = [
            'presence_absence1\tpresence_absence1\t1\t0\t539\t292\tcluster_name\t117\t117\t92.31\tcluster_name.l15.c30.ctg.1\t1104\t20.0\t0\t.\tp\t.\t0\tR25_A26del\tDEL\t73\t73\tA\t553\t553\tA\t27\tA\t27\t.\tGeneric description of presence_absence1',
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)


    def test_full_run_insert_codon(self):
        '''Test complete run where there is a inserted codon'''
        fasta_in = os.path.join(data_dir, 'cluster_test_full_run_insert_codon.fa')
        tsv_in = os.path.join(data_dir, 'cluster_test_full_run_insert_codon.tsv')
        refdata = reference_data.ReferenceData([fasta_in], [tsv_in])
        tmpdir = 'tmp.cluster_test_full_insert_codon'
        common.rmtree(tmpdir)
        shutil.copytree(os.path.join(data_dir, 'cluster_test_full_run_insert_codon'), tmpdir)
        c = cluster.Cluster(tmpdir, 'cluster_name', refdata, total_reads=292, total_reads_bases=20900)
        c.run()

        expected = [
            'presence_absence1\tpresence_absence1\t1\t0\t539\t292\tcluster_name\t108\t108\t92.31\tcluster_name.l15.c30.ctg.1\t1115\t19.9\t0\t.\tp\t.\t0\tS25_M26insELI\tINS\t73\t73\tA\t554\t554\tG\t24\tG\t24\t.\tGeneric description of presence_absence1'
        ]
        self.assertEqual(expected, c.report_lines)
        common.rmtree(tmpdir)



import unittest
import os
from ariba import flag, summary_cluster, summary_cluster_variant

modules_dir = os.path.dirname(os.path.abspath(summary_cluster.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestSummaryCluster(unittest.TestCase):
    def test_line2dict(self):
        '''Test _line2dict'''
        line = 'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t120\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:var_group1:ref has wild type, foo bar\tsome free text'

        expected = {
            'ariba_ref_name': 'ariba_refname',
            'ref_name': 'refname',
            'gene': '1',
            'var_only' : '0',
            'flag': flag.Flag(19),
            'reads': 78,
            'cluster': 'cluster',
            'ref_len': 120,
            'ref_base_assembled': 120,
            'pc_ident': 98.33,
            'ctg': 'ctg_name',
            'ctg_len': 279,
            'ctg_cov': 24.4,
            'known_var': '1',
            'var_type': 'SNP',
            'var_seq_type': 'n',
            'known_var_change': 'A14T',
            'has_known_var': '1',
            'ref_ctg_change': 'A14T',
            'ref_ctg_effect': 'SNP',
            'ref_start': 13,
            'ref_end': 13,
            'ref_nt': 'A',
            'ctg_start': 84,
            'ctg_end': 84,
            'ctg_nt': 'T',
            'smtls_total_depth': '17',
            'smtls_nts': 'T',
            'smtls_nts_depth': '17',
            'var_description': 'noncoding1:1:0:A14T:var_group1:ref has wild type, foo bar',
            'var_group': 'var_group1',
            'free_text': 'some free text'
        }

        self.assertEqual(summary_cluster.SummaryCluster.line2dict(line), expected)


    def test_add_data_dict(self):
        '''Test add_data_dict'''
        cluster = summary_cluster.SummaryCluster()
        self.assertTrue(cluster.name is None)
        line1 = 'ariba_refname1\trefname\t1\t0\t19\t78\tcluster\t120\t120\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text'
        line2 = 'ariba_refname1\trefname\t1\t0\t19\t78\tcluster2\t120\t120\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id2:ref has wild type, foo bar\tsome free text'
        line3 = 'ariba_refname2\trefname2\t1\t0\t19\t78\tcluster\t120\t120\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id3:ref has wild type, foo bar\tsome free text'
        data_dict1 = summary_cluster.SummaryCluster.line2dict(line1)
        data_dict2 = summary_cluster.SummaryCluster.line2dict(line2)
        data_dict3 = summary_cluster.SummaryCluster.line2dict(line3)
        cluster.add_data_dict(data_dict1)
        self.assertEqual(cluster.name, data_dict1['cluster'])
        self.assertEqual(cluster.data,[data_dict1])
        with self.assertRaises(summary_cluster.Error):
            cluster.add_data_dict(data_dict2)

        with self.assertRaises(summary_cluster.Error):
            cluster.add_data_dict(data_dict3)


    def test_has_any_part_of_ref_assembled(self):
        '''Test _has_any_part_of_ref_assembled'''
        line1 = 'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t.\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text'
        line2 = 'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t0\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text'
        line3 = 'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text'
        data_dict1 = summary_cluster.SummaryCluster.line2dict(line1)
        data_dict2 = summary_cluster.SummaryCluster.line2dict(line2)
        data_dict3 = summary_cluster.SummaryCluster.line2dict(line3)
        cluster = summary_cluster.SummaryCluster()
        cluster.add_data_dict(data_dict1)
        self.assertFalse(cluster._has_any_part_of_ref_assembled())
        cluster.add_data_dict(data_dict2)
        self.assertFalse(cluster._has_any_part_of_ref_assembled())
        cluster.add_data_dict(data_dict3)
        self.assertTrue(cluster._has_any_part_of_ref_assembled())


    def test_pc_id_and_read_depth_of_longest(self):
        '''Test _pc_id_and_read_depth_of_longest'''
        cluster = summary_cluster.SummaryCluster()
        self.assertTrue(cluster.name is None)
        line1 = 'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t42.2\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text'
        line2 = 'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t119\t98.20\tctg_name2\t279\t42.42\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text'
        line3 = 'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t114\t98.32\tctg_name3\t279\t42.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text'
        data_dict1 = summary_cluster.SummaryCluster.line2dict(line1)
        data_dict2 = summary_cluster.SummaryCluster.line2dict(line2)
        data_dict3 = summary_cluster.SummaryCluster.line2dict(line3)
        cluster.add_data_dict(data_dict1)
        cluster.add_data_dict(data_dict2)
        cluster.add_data_dict(data_dict3)
        self.assertEqual((98.2, 42.42), cluster._pc_id_and_read_depth_of_longest())


    def test_to_cluster_summary_number(self):
        '''Test _to_cluster_summary_assembled'''
        line = 'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text'
        data_dict = summary_cluster.SummaryCluster.line2dict(line)

        tests = [
            ('0', 0, 'partial'),
            ('0', 64, 'no'),
            ('0', 1024, 'no'),
            ('0', 1, 'fragmented'),
            ('0', 3, 'yes_nonunique'),
            ('0', 19, 'yes'),
            ('0', 23, 'yes_nonunique'),
            ('0', 51, 'yes_nonunique'),
            ('0', 147, 'yes_nonunique'),
            ('0', 275, 'yes_nonunique'),
            ('1', 0, 'partial'),
            ('1', 64, 'no'),
            ('1', 1024, 'no'),
            ('1', 1, 'fragmented'),
            ('1', 11, 'yes_nonunique'),
            ('1', 27, 'yes'),
            ('1', 29, 'fragmented'),
            ('1', 59, 'yes_nonunique'),
            ('1', 155, 'yes_nonunique'),
            ('1', 283, 'yes_nonunique'),
        ]

        for gene, f, expected in tests:
            cluster = summary_cluster.SummaryCluster()
            data_dict['gene'] = gene
            data_dict['flag'] = flag.Flag(f)
            cluster.add_data_dict(data_dict)
            self.assertEqual(expected, cluster._to_cluster_summary_assembled())
            if expected == 'partial':
                original_number = cluster.data[0]['ref_base_assembled']
                cluster.data[0]['ref_base_assembled'] = 0
                self.assertEqual('no', cluster._to_cluster_summary_assembled())
                cluster.data[0]['ref_base_assembled'] = original_number


    def test_has_known_variant(self):
        '''Test _has_known_variant'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\t.\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tMULTIPLE\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tINDELS\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_23S.rDNA_WHO_F_01358c\t23S.rDNA_WHO_F_01358c\t0\t1\t659\t4168\t23S\t2890\t2890\t99.86\t23S.scaffold.1\t3628\t344.0\t1\tSNP\tn\tC2597T\t0\t.\t.\t2597\t2597\tC\t2928\t2928\tC\t410\tC,T\t70,30\t23S.rDNA_WHO_F_01358c:0:1:C2597T:.:E coli C2611T\t.',
        ]

        dicts = [summary_cluster.SummaryCluster.line2dict(x) for x in lines]
        expected = ['yes', 'no', 'no', 'no', 'no', 'het']
        assert len(dicts) == len(expected)

        for i in range(len(dicts)):
            self.assertEqual(expected[i], summary_cluster.SummaryCluster._has_known_variant(dicts[i]))


    def test_has_any_known_variant(self):
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\t.\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tMULTIPLE\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tINDELS\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_23S.rDNA_WHO_F_01358c\t23S.rDNA_WHO_F_01358c\t0\t1\t659\t4168\t23S\t2890\t2890\t99.86\t23S.scaffold.1\t3628\t344.0\t1\tSNP\tn\tC2597T\t0\t.\t.\t2597\t2597\tC\t2928\t2928\tC\t410\tC,T\t70,30\t23S.rDNA_WHO_F_01358c:0:1:C2597T:.:E coli C2611T\t.',
        ]

        expected = ['yes', 'no', 'no', 'no', 'no', 'het']
        assert len(lines) == len(expected)

        for i in range(len(lines)):
            data_dict = summary_cluster.SummaryCluster.line2dict(lines[i])
            cluster = summary_cluster.SummaryCluster()
            cluster.add_data_dict(data_dict)
            self.assertEqual(expected[i], cluster._has_any_known_variant())


    def test_has_nonsynonymous(self):
        '''Test _has_nonsynonymous'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSYN\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\t.\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tMULTIPLE\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tINDELS\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_23S.rDNA_WHO_F_01358c\t23S.rDNA_WHO_F_01358c\t0\t1\t659\t4168\t23S\t2890\t2890\t99.86\t23S.scaffold.1\t3628\t344.0\t0\tHET\t.\t.\t.\t.\t.\t2597\t2597\tC\t2928\t2928\tC\t410\tC,T\t70,30\t.\t.',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA62T\t1\tA62T\tSNP\t62\t62\tA\t84\t84\tA\t40\tA,T\t10,30\tnon_coding1:0:0:A62T:id2:foo_bar\tspam eggs',
        ]

        dicts = [summary_cluster.SummaryCluster.line2dict(x) for x in lines]
        expected = ['no', 'yes', 'no', 'yes', 'yes', 'yes', 'het', 'het']
        assert len(dicts) == len(expected)

        for i in range(len(dicts)):
            self.assertEqual(expected[i], summary_cluster.SummaryCluster._has_nonsynonymous(dicts[i]))


    def test_has_any_nonsynonymous(self):
        '''Test _has_any_nonsynonymous'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSYN\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:N_ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\t.\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\t.\t.\t.\tMULTIPLE\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_23S.rDNA_WHO_F_01358c\t23S.rDNA_WHO_F_01358c\t0\t1\t659\t4168\t23S\t2890\t2890\t99.86\t23S.scaffold.1\t3628\t344.0\t0\tHET\t.\t.\t.\t.\t.\t2597\t2597\tC\t2928\t2928\tC\t410\tC,T\t70,30\t.\t.',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA62T\t1\tA62T\tSNP\t62\t62\tA\t84\t84\tA\t40\tA,T\t10,30\tnon_coding1:0:0:A62T:id2:foo_bar\tspam eggs',
        ]

        expected = ['no', 'yes', 'no', 'yes', 'yes', 'het', 'het']
        assert len(lines) == len(expected)

        for i in range(len(lines)):
            data_dict = summary_cluster.SummaryCluster.line2dict(lines[i])
            cluster = summary_cluster.SummaryCluster()
            cluster.add_data_dict(data_dict)
            self.assertEqual(expected[i], cluster._has_any_nonsynonymous())


    def test_has_novel_nonsynonymous(self):
        '''Test _has_novel_nonsynonymous'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\t.\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tMULTIPLE\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tINDELS\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_23S.rDNA_WHO_F_01358c\t23S.rDNA_WHO_F_01358c\t0\t1\t659\t4168\t23S\t2890\t2890\t99.86\t23S.scaffold.1\t3628\t344.0\t1\tSNP\tn\tC2597T\t0\t.\t.\t2597\t2597\tC\t2928\t2928\tC\t410\tC,T\t70,30\t23S.rDNA_WHO_F_01358c:0:1:C2597T:.:E coli C2611T\t.',
            'ariba_23S.rDNA_WHO_F_01358c\t23S.rDNA_WHO_F_01358c\t0\t1\t659\t4168\t23S\t2890\t2890\t99.86\t23S.scaffold.1\t3628\t344.0\t0\tHET\t.\t.\t.\t.\t.\t2597\t2597\tC\t2928\t2928\tC\t410\tC,T\t70,30\t.\t.',
        ]

        dicts = [summary_cluster.SummaryCluster.line2dict(x) for x in lines]
        expected = ['no', 'no', 'yes', 'yes', 'yes', 'no', 'het']
        assert len(dicts) == len(expected)

        for i in range(len(dicts)):
            self.assertEqual(expected[i], summary_cluster.SummaryCluster._has_novel_nonsynonymous(dicts[i]))


    def test_has_any_novel_nonsynonymous(self):
        '''Test _has_any_novel_nonsynonymous'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\t.\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\t.\tsome free text',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tMULTIPLE\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_refname\trefname\t1\t0\t528\t2814\tcluster\t1188\t1009\t90.49\tctg_name\t2470\t141.8\t0\t.\tp\t.\t0\t.\tINDELS\t594\t594\tC;T\t1195\t1195\t.;C\t207;204\t.;.\t207;204\t.\t.',
            'ariba_23S.rDNA_WHO_F_01358c\t23S.rDNA_WHO_F_01358c\t0\t1\t659\t4168\t23S\t2890\t2890\t99.86\t23S.scaffold.1\t3628\t344.0\t1\tSNP\tn\tC2597T\t0\t.\t.\t2597\t2597\tC\t2928\t2928\tC\t410\tC,T\t70,30\t23S.rDNA_WHO_F_01358c:0:1:C2597T:.:E coli C2611T\t.',
            'ariba_23S.rDNA_WHO_F_01358c\t23S.rDNA_WHO_F_01358c\t0\t1\t659\t4168\t23S\t2890\t2890\t99.86\t23S.scaffold.1\t3628\t344.0\t0\tHET\t.\t.\t.\t.\t.\t2597\t2597\tC\t2928\t2928\tC\t410\tC,T\t70,30\t.\t.',
        ]

        expected = ['no', 'no', 'yes', 'yes', 'yes', 'no', 'het']
        assert len(lines) == len(expected)

        for i in range(len(lines)):
            data_dict = summary_cluster.SummaryCluster.line2dict(lines[i])
            cluster = summary_cluster.SummaryCluster()
            cluster.add_data_dict(data_dict)
            self.assertEqual(expected[i], cluster._has_any_novel_nonsynonymous())


    def test_to_cluster_summary_has_known_nonsynonymous(self):
        '''Test _to_cluster_summary_has_known_nonsynonymous'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSYN\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\t.\tn\t.\t.\t.\tMULTIPLE\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
        ]

        expected = ['yes', 'yes', 'no', 'no', 'no']

        for i in range(len(lines)):
            data_dict = summary_cluster.SummaryCluster.line2dict(lines[i])
            cluster = summary_cluster.SummaryCluster()
            cluster.add_data_dict(data_dict)
            for assembled_summary in ['yes', 'fragmented', 'yes_nonunique']:
                self.assertEqual(expected[i], cluster._to_cluster_summary_has_known_nonsynonymous(assembled_summary))
            self.assertEqual('NA', cluster._to_cluster_summary_has_known_nonsynonymous('no'))


    def test_to_cluster_summary_has_novel_nonsynonymous(self):
        '''Test _to_cluster_summary_has_novel_nonsynonymous'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSYN\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\t.\tn\t.\t.\t.\tMULTIPLE\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
        ]

        expected = ['no', 'no', 'no', 'yes', 'yes']

        for i in range(len(lines)):
            data_dict = summary_cluster.SummaryCluster.line2dict(lines[i])
            cluster = summary_cluster.SummaryCluster()
            cluster.add_data_dict(data_dict)
            for assembled_summary in ['yes', 'fragmented', 'yes_nonunique']:
                self.assertEqual(expected[i], cluster._to_cluster_summary_has_novel_nonsynonymous(assembled_summary))
            self.assertEqual('NA', cluster._to_cluster_summary_has_novel_nonsynonymous('no'))


    def test_to_cluster_summary_has_nonsynonymous(self):
        '''Test _to_cluster_summary_has_nonsynonymous'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSYN\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\t.\tn\t.\t.\t.\tMULTIPLE\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
        ]

        expected = ['no', 'yes', 'no', 'yes', 'yes']

        for i in range(len(lines)):
            data_dict = summary_cluster.SummaryCluster.line2dict(lines[i])
            cluster = summary_cluster.SummaryCluster()
            cluster.add_data_dict(data_dict)
            for assembled_summary in ['yes', 'fragmented', 'yes_nonunique']:
                self.assertEqual(expected[i], cluster._to_cluster_summary_has_nonsynonymous(assembled_summary))
            self.assertEqual('NA', cluster._to_cluster_summary_has_nonsynonymous('no'))


    def test_get_known_noncoding_het_snp(self):
        '''Test _get_known_noncoding_het_snp'''
        d = {'gene': '1'}
        self.assertEqual(None, summary_cluster.SummaryCluster._get_known_noncoding_het_snp(d))
        d['gene'] = '0'
        d['known_var'] = '0'
        self.assertEqual(None, summary_cluster.SummaryCluster._get_known_noncoding_het_snp(d))

        d['known_var'] = '1'
        d['ref_ctg_effect'] = 'not_a_snp'
        self.assertEqual(None, summary_cluster.SummaryCluster._get_known_noncoding_het_snp(d))

        d['ref_ctg_effect'] = 'SNP'
        d['smtls_nts'] = '.'
        self.assertEqual(None, summary_cluster.SummaryCluster._get_known_noncoding_het_snp(d))

        d['smtls_nts'] = 'A;G;T'
        self.assertEqual(None, summary_cluster.SummaryCluster._get_known_noncoding_het_snp(d))

        d['known_var_change'] = 'A42T'
        d['ctg_nt'] = 'A'
        d['smtls_nts'] = 'A,T'
        d['smtls_nts_depth'] = '52,48'
        self.assertEqual(('A42T', 48.0), summary_cluster.SummaryCluster._get_known_noncoding_het_snp(d))


    def test_get_nonsynonymous_var(self):
        '''Test _get_nonsynonymous_var'''
        d = {
            'ref_name': 'ref',
            'gene': '1',
            'var_type': '.',
            'known_var_change': '.',
            'has_known_var': '.',
            'known_var': '0',
            'ref_ctg_change': '.',
            'ref_ctg_effect': '.',
            'var_seq_type': '.',
            'var_group': '.',
        }

        self.assertEqual(None, summary_cluster.SummaryCluster._get_nonsynonymous_var(d))

        d['var_type'] = 'p'
        d['known_var'] = '1'
        d['has_known_var'] = '1'
        with self.assertRaises(summary_cluster_variant.Error):
            summary_cluster.SummaryCluster._get_nonsynonymous_var(d)

        d['known_var_change'] = 'I42L'
        self.assertEqual(('ref', 'I42L', 'ungrouped', None), summary_cluster.SummaryCluster._get_nonsynonymous_var(d))

        d['var_group'] = 'vgroup'
        self.assertEqual(('ref', 'I42L', 'grouped', 'vgroup'), summary_cluster.SummaryCluster._get_nonsynonymous_var(d))
        d['var_group'] = '.'

        d['ref_ctg_change'] = 'P43Q'
        with self.assertRaises(summary_cluster_variant.Error):
            summary_cluster.SummaryCluster._get_nonsynonymous_var(d)

        d['known_var_change'] = '.'
        self.assertEqual(('ref', 'P43Q', 'novel', None), summary_cluster.SummaryCluster._get_nonsynonymous_var(d))

        d['ref_ctg_change'] = '.'
        with self.assertRaises(summary_cluster_variant.Error):
            summary_cluster.SummaryCluster._get_nonsynonymous_var(d)

        d['ref_ctg_effect'] = 'MULTIPLE'
        self.assertEqual(('ref', 'MULTIPLE', 'novel', None), summary_cluster.SummaryCluster._get_nonsynonymous_var(d))


    def test_has_match(self):
        '''Test _has_match'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSYN\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t1\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSYN\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:1:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t1\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:1:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tp\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tp\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tp\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t1\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tp\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:1:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t1\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tp\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:1:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t1\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tp\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:1:A14T:id1:ref has wild type, foo bar\tsome free text',
        ]

        expected = ['yes', 'yes', 'yes', 'no', 'yes', 'yes', 'yes', 'yes', 'no', 'no']

        for i in range(len(lines)):
            data_dict = summary_cluster.SummaryCluster.line2dict(lines[i])
            cluster = summary_cluster.SummaryCluster()
            cluster.add_data_dict(data_dict)
            for assembled_summary in ['yes', 'yes_nonunique']:
                self.assertEqual(expected[i], cluster._has_match(assembled_summary))
            for assembled_summary in ['no', 'fragmented']:
                self.assertEqual('no', cluster._has_match(assembled_summary))


    def test_has_var_groups(self):
        '''Test has_var_groups'''
        lines = [
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSYN\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id1:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t0\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tn\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:0:0:A14T:id2:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tp\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id3:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tp\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id4:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t0\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tp\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:0:A14T:id5:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t1\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tp\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:1:A14T:id6:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t1\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tp\tA14T\t0\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:1:A14T:id7:ref has wild type, foo bar\tsome free text',
            'ariba_refname\trefname\t1\t1\t19\t78\tcluster\t120\t100\t98.33\tctg_name\t279\t24.4\t0\tSNP\tp\tA14T\t.\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnoncoding1:1:1:A14T:id7:ref has wild type, foo bar\tsome free text',
        ]
        dicts = [summary_cluster.SummaryCluster.line2dict(line) for line in lines]
        cluster = summary_cluster.SummaryCluster()
        for d in dicts:
            cluster.add_data_dict(d)
        got = cluster.has_var_groups()
        expected = {'id1', 'id3', 'id6'}
        self.assertEqual(expected, got)


    def test_column_summary_data(self):
        '''Test column_summary_data'''
        line1 = 'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnon_coding1:n:A14T:id1:foo_bar\tspam eggs'
        line2 = 'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t95\t98.42\tctg_name\t279\t24.4\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\tsome free text'

        data_dict1 = summary_cluster.SummaryCluster.line2dict(line1)
        data_dict2 = summary_cluster.SummaryCluster.line2dict(line2)
        cluster = summary_cluster.SummaryCluster()
        cluster.add_data_dict(data_dict1)
        cluster.add_data_dict(data_dict2)
        expected = {
            'assembled': 'yes',
            'match': 'yes',
            'ref_seq': 'ref1',
            'novel_var': 'no',
            'known_var': 'yes',
            'pct_id': '98.33',
            'ctg_cov': '24.4',
        }
        got = cluster.column_summary_data()
        self.assertEqual(expected, got)


    def test_non_synon_variants(self):
        '''Test non_synon_variants'''
        line1 = 'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnon_coding1:0:0:A14T:id1:foo_bar\tspam eggs'
        line2 = 'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t95\t98.42\tctg_name\t279\t24.4\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\tsome free text'

        data_dict1 = summary_cluster.SummaryCluster.line2dict(line1)
        data_dict2 = summary_cluster.SummaryCluster.line2dict(line2)
        cluster = summary_cluster.SummaryCluster()
        cluster.add_data_dict(data_dict1)
        cluster.add_data_dict(data_dict2)
        got = cluster.non_synon_variants()
        expected = {('ref1', 'A14T', 'grouped', 'id1')}
        self.assertEqual(expected, got)


    def test_known_noncoding_het_snps(self):
        '''test known_noncoding_het_snps'''
        lines = [
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnon_coding1:0:0:A14T:id1:foo_bar\tspam eggs',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA42T\t1\tA42T\tSNP\t42\t42\tA\t84\t84\tT\t40\tT,A\t10,30\tnon_coding1:0:0:A42T:id1:foo_bar\tspam eggs',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA62T\t1\tA62T\tSNP\t62\t62\tA\t84\t84\tA\t40\tA,T\t10,30\tnon_coding1:0:0:A62T:id2:foo_bar\tspam eggs',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA82T\t1\tA82T\tSNP\t82\t82\tA\t84\t84\tA\t100\tA,T,G\t10,40,50\tnon_coding1:0:0:A82T:.:foo_bar\tspam eggs'
        ]

        cluster = summary_cluster.SummaryCluster()
        for line in lines:
            cluster.add_data_dict(summary_cluster.SummaryCluster.line2dict(line))
        got = cluster.known_noncoding_het_snps()
        expected = {
            '.': {'A82T': 40.0},
            'id1': {'A42T': 25.0, 'A14T': 100.0},
            'id2': {'A62T': 75.0},
        }
        self.assertEqual(expected, got)


    def test_get_all_nonsynon_variants_set(self):
        '''test _get_all_nonsynon_variants_set'''
        lines = [
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t95\t98.42\tctg_name\t279\t24.4\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\tsome free text',
            'ariba_ref1\tref1\t1\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tI14L\t1\tI14L\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnon_coding1:0:0:I14L:.:foo_bar\tspam eggs',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t40\tT,A\t10,30\tnon_coding1:0:0:A14T:id1:foo_bar\tspam eggs',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t40\tT,A,G\t20,10,10\tnon_coding1:0:0:A14T:id1:foo_bar\tspam eggs',
        ]

        data_dicts = [summary_cluster.SummaryCluster.line2dict(x) for x in lines]

        cluster_vars = [summary_cluster_variant.SummaryClusterVariant(x) for x in data_dicts]
        expected = {x for x in cluster_vars if x.has_nonsynon}
        got = summary_cluster.SummaryCluster._get_all_nonsynon_variants_set(data_dicts)
        self.assertEqual(expected, got)


    def test_gather_data(self):
        '''test gather_data'''
        lines = [
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t95\t98.42\tctg_name\t279\t24.4\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\t.\tsome free text',
            'ariba_ref1\tref1\t1\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tI14L\t1\tI14L\tSNP\t13\t13\tA\t84\t84\tT\t17\tT\t17\tnon_coding1:0:0:I14L:.:foo_bar\tspam eggs',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t40\tT,A\t10,30\tnon_coding1:0:0:A14T:id1:foo_bar\tspam eggs',
            'ariba_ref1\tref1\t0\t0\t531\t78\tcluster1\t120\t100\t98.33\tctg_name\t279\t24.4\t1\tSNP\tn\tA14T\t1\tA14T\tSNP\t13\t13\tA\t84\t84\tT\t40\tT,A,G\t20,10,10\tnon_coding1:0:0:A14T:id1:foo_bar\tspam eggs',
        ]

        data_dicts = [summary_cluster.SummaryCluster.line2dict(x) for x in lines]
        cluster = summary_cluster.SummaryCluster()
        for data_dict in data_dicts:
            cluster.add_data_dict(data_dict)

        cluster.gather_data()
        expected_summary = {
            'assembled': 'yes',
            'match': 'yes',
            'ref_seq': 'ref1',
            'pct_id': '98.33',
            'ctg_cov': '24.4',
            'known_var': 'yes',
            'novel_var': 'no',
        }
        self.assertEqual(expected_summary, cluster.summary)

        cluster_vars = [summary_cluster_variant.SummaryClusterVariant(x) for x in data_dicts]
        expected_variants = {x for x in cluster_vars if x.has_nonsynon}
        self.assertEqual(expected_variants, cluster.variants)




import unittest
import os
from ariba import histogram

modules_dir = os.path.dirname(os.path.abspath(histogram.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestCluster(unittest.TestCase):
    def test_to_bin(self):
        '''Test _to_bin'''
        h = histogram.Histogram(3)
        tests = [
            (0, 0),
            (1, 0),
            (2, 0),
            (3, 3),
            (4, 3),
            (5, 3),
            (6, 6),
        ]

        for t in tests:
            self.assertEqual(h._to_bin(t[0]), t[1])


    def test_len(self):
        '''Test __len__'''
        h = histogram.Histogram(1)
        self.assertEqual(0, len(h))
        h.add(1)
        self.assertEqual(1, len(h))
        h.add(1)
        self.assertEqual(2, len(h))
        h.add(2)
        self.assertEqual(3, len(h))


    def test_add(self):
        '''Test add'''
        h = histogram.Histogram(3)
        h.add(4)
        self.assertEqual({3:1}, h.bins)
        h.add(4)
        self.assertEqual({3:2}, h.bins)
        h.add(42)
        self.assertEqual({3:2, 42:1}, h.bins)
        h.add(42, count=11)
        self.assertEqual({3:2, 42:12}, h.bins)


    def test_stats(self):
        '''Test stats'''
        h = histogram.Histogram(1)
        for i in range(10):
            h.add(i+1)
        h.bins[3] = 3
        h.bins[4] = 3
        h.bins[5] = 5
        h.bins[6] = 3
        h.bins[7] = 2
        h.bins[8] = 2
        self.assertEqual((2.5, 5.5, 10.5, 0.91), h.stats())



import os
import pickle
from ariba import reference_data


class Error (Exception): pass

class RefdataQuery:
    def __init__(self,
        prepareref_dir
    ):
        if not os.path.exists(prepareref_dir):
            raise Error('Error querying refdata. Input directory "' + prepareref_dir + '" not found.')

        self.prepareref_dir = prepareref_dir
        self.refdata_fa = os.path.join(self.prepareref_dir, '02.cdhit.all.fa')
        self.refdata_tsv = os.path.join(self.prepareref_dir, '01.filter.check_metadata.tsv')
        self.clusters_pickle = os.path.join(self.prepareref_dir, '02.cdhit.clusters.pickle')


    @staticmethod
    def _load_clusters(pickle_file):
        with open(pickle_file, 'rb') as f:
            clusters = pickle.load(f)
        return clusters


    @staticmethod
    def _seq2cluster(clusters, seqname):
        for cluster in clusters:
            if seqname in clusters[cluster]:
                return cluster
        return None


    def _cluster2seqs(self, cluster_name):
        clusters = self._load_clusters(self.clusters_pickle)
        if cluster_name not in clusters:
            return ['Cluster name "' + cluster_name + '" not found']
        else:
            return ['Sequences belonging to cluster ' + cluster_name + ':'] + sorted(list(clusters[cluster_name]))


    def _seqinfo(self, seqname):
        refdata = reference_data.ReferenceData([self.refdata_fa], [self.refdata_tsv])
        if seqname not in refdata.sequences:
            return ['Sequence "' + seqname + '" not found']

        assert seqname in refdata.metadata

        clusters = RefdataQuery._load_clusters(self.clusters_pickle)
        cluster = RefdataQuery._seq2cluster(clusters, seqname)
        assert cluster is not None

        if refdata.metadata[seqname]['seq_type'] == 'p':
            is_gene = '1'
            var_dict = refdata.metadata[seqname]['p']
        else:
            is_gene = '0'
            var_dict = refdata.metadata[seqname]['n']

        description_lines = ['Description\t' + x.free_text for x in sorted(list(refdata.metadata[seqname]['.'])) if x.free_text != '.']

        var_lines = []
        for position in sorted(var_dict):
            for seq_meta in sorted(var_dict[position]):
                ident = '.' if seq_meta.variant.identifier is None else seq_meta.variant.identifier
                var_lines.append('\t'.join(['Variant', str(seq_meta.variant), ident, seq_meta.free_text]))

        return [
            'Name\t' + seqname,
            'Is gene\t' + is_gene,
            'Variant only\t' + ('1' if refdata.metadata[seqname]['variant_only'] else '0'),
            'Cluster\t' + cluster,
        ] + description_lines + var_lines + ['Sequence\t' + refdata.sequences[seqname].seq]


    def query(self, query_type, query_string):
        queries = {
            'cluster': self._cluster2seqs,
            'seq': self._seqinfo,
        }

        if query_type not in queries:
            raise Error('Unknown query type "' + query_type + '". Choices are:\n' + ','.join(sorted(queries.keys())) + '\n')

        lines = queries[query_type](query_string)
        print(*lines, sep='\n')




import unittest
import filecmp
import os
from ariba import faidx, external_progs

modules_dir = os.path.dirname(os.path.abspath(faidx.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')
extern_progs = external_progs.ExternalProgs()


class TestFaidx(unittest.TestCase):
    def test_write_fa_subset(self):
        '''test write_fa_subset'''
        infile = os.path.join(data_dir, 'faidx_test_write_fa_subset.in.fa')
        expected = os.path.join(data_dir, 'faidx_test_write_fa_subset.out.fa')
        tmpfile = 'tmp.test_write_fa_subset.out.fa'
        faidx.write_fa_subset(['seq1', 'seq3', 'seq4'], infile, tmpfile)
        self.assertTrue(filecmp.cmp(expected, tmpfile, shallow=False))
        os.unlink(tmpfile)



import unittest
import os
import filecmp
import pyfastaq
from ariba import aln_to_metadata, sequence_variant

modules_dir = os.path.dirname(os.path.abspath(aln_to_metadata.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


class TestAlnToMetadata(unittest.TestCase):
    def test_load_aln_file(self):
        '''test _load_aln_file'''
        aln_file = os.path.join(data_dir, 'aln_to_metadata_load_aln_file.in.fa')
        expected = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'ABC-DE'),
            'seq2': pyfastaq.sequences.Fasta('seq2', 'ABCQDE'),
        }
        got = aln_to_metadata.AlnToMetadata._load_aln_file(aln_file)
        self.assertEqual(expected, got)


    def test_load_vars_file_good_file(self):
        '''test _load_vars_file good input file'''
        infile = os.path.join(data_dir, 'aln_to_metadata_load_vars_file_good.tsv')
        variant1 = sequence_variant.Variant('p', 'A42B', 'id1')
        variant2 = sequence_variant.Variant('p', 'C43D', 'id2')
        variant3 = sequence_variant.Variant('p', 'E100F', 'id3')
        expected = {
            'seq1': [(variant1, 'description 1')],
            'seq2': [(variant2, 'description 2'), (variant3, 'description 3')]
        }
        got = aln_to_metadata.AlnToMetadata._load_vars_file(infile, True)
        self.assertEqual(expected, got)


    def test_load_vars_bad_files(self):
        '''test _load_vars_file bad input files'''
        infiles = [
            os.path.join(data_dir, 'aln_to_metadata_load_vars_file_bad.1.tsv'),
            os.path.join(data_dir, 'aln_to_metadata_load_vars_file_bad.2.tsv')
        ]

        for infile in infiles:
            with self.assertRaises(aln_to_metadata.Error):
                aln_to_metadata.AlnToMetadata._load_vars_file(infile, True)


    def test_make_unpadded_seqs(self):
        '''test _make_unpadded_seqs'''
        padded = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'acg---t'),
            'seq2': pyfastaq.sequences.Fasta('seq2', '---a-cgt-'),
        }
        expected = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'acgt'),
            'seq2': pyfastaq.sequences.Fasta('seq2', 'acgt'),
        }
        got = aln_to_metadata.AlnToMetadata._make_unpadded_seqs(padded)
        self.assertEqual(expected, got)


    def test_check_seq_lengths_same(self):
        '''test _check_seq_lengths_same'''
        seqs = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'acgt'),
            'seq2': pyfastaq.sequences.Fasta('seq2', 'acgt'),
        }

        self.assertTrue(aln_to_metadata.AlnToMetadata._check_seq_lengths_same(seqs))
        seqs['seq1'].seq = 'a'
        with self.assertRaises(aln_to_metadata.Error):
            aln_to_metadata.AlnToMetadata._check_seq_lengths_same(seqs)


    def test_insertion_coords(self):
        '''test _insertion_coords'''
        tests = [
            ('acgt', []),
            ('-a', [pyfastaq.intervals.Interval(0, 0)]),
            ('a---cgt--', [pyfastaq.intervals.Interval(1, 3), pyfastaq.intervals.Interval(7, 8)]),
        ]

        for seq, expected in tests:
            fa = pyfastaq.sequences.Fasta('x', seq)
            got = aln_to_metadata.AlnToMetadata._insertion_coords(fa)
            self.assertEqual(expected, got)


    def test_make_unpadded_insertion_coords(self):
        '''test _make_unpadded_insertion_coords'''
        seqs = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'acgt'),
            'seq2': pyfastaq.sequences.Fasta('seq2', 'ac-gt'),
            'seq3': pyfastaq.sequences.Fasta('seq3', '--acg-t'),
        }

        expected = {
            'seq1': [],
            'seq2': [pyfastaq.intervals.Interval(2, 2)],
            'seq3': [pyfastaq.intervals.Interval(0, 1), pyfastaq.intervals.Interval(5, 5)],

        }
        got = aln_to_metadata.AlnToMetadata._make_unpadded_insertion_coords(seqs)
        self.assertEqual(expected, got)


    def test_check_insertion_coords(self):
        '''test _check_insertion_coords'''
        seq = pyfastaq.sequences.Fasta('name', 'AAA---GGG------TTT---')
        self.assertTrue(aln_to_metadata.AlnToMetadata._check_insertion_coords(seq))

        bad_seqs = [
            pyfastaq.sequences.Fasta('name', 'AAA--GGG'),  # bad length
            pyfastaq.sequences.Fasta('name', 'A---AA'),  # bad start position
            pyfastaq.sequences.Fasta('name', 'AA---AA'), # bad start position
        ]

        for seq in bad_seqs:
            with self.assertRaises(aln_to_metadata.Error):
                aln_to_metadata.AlnToMetadata._check_insertion_coords(seq)


    def test_check_coding_seq(self):
        '''test _check_coding_seq'''
        seq = pyfastaq.sequences.Fasta('name', 'ATGCTTTAG')
        self.assertTrue(aln_to_metadata.AlnToMetadata._check_coding_seq(seq))

        bad_seqs = [
            pyfastaq.sequences.Fasta('name', 'TTGCTTAG'), # length not a mutliple of 3
            pyfastaq.sequences.Fasta('name', 'TTTCTTTAG'), # no start codon
            pyfastaq.sequences.Fasta('name', 'ATGTAGCTTTAG'), # stop codon in middle
            pyfastaq.sequences.Fasta('name', 'TTGCTTTTT'), # no stop at end
        ]

        for seq in bad_seqs:
            with self.assertRaises(aln_to_metadata.Error):
                aln_to_metadata.AlnToMetadata._check_coding_seq(seq)


    def test_check_sequences_non_coding(self):
        '''test _check_sequences with noncoding seqs'''
        padded_sequences = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'AC-T')
        }

        unpadded_sequences = aln_to_metadata.AlnToMetadata._make_unpadded_seqs(padded_sequences)
        self.assertTrue(aln_to_metadata.AlnToMetadata._check_sequences(padded_sequences, unpadded_sequences, False))
        padded_sequences['seq2'] = pyfastaq.sequences.Fasta('seq2', 'AC-')
        unpadded_sequences = aln_to_metadata.AlnToMetadata._make_unpadded_seqs(padded_sequences)
        with self.assertRaises(aln_to_metadata.Error):
            aln_to_metadata.AlnToMetadata._check_sequences(padded_sequences, unpadded_sequences, False)


    def test_check_sequences_coding(self):
        '''test _check_sequences with coding seqs'''
        padded_sequences = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'ATGCTTTAG'),
            'seq2': pyfastaq.sequences.Fasta('seq2', 'ATG---TAG')
        }

        unpadded_sequences = aln_to_metadata.AlnToMetadata._make_unpadded_seqs(padded_sequences)

        self.assertTrue(aln_to_metadata.AlnToMetadata._check_sequences(padded_sequences, unpadded_sequences, True))

        bad_seqs = [
            'ATGCTTAG', # length not a mutliple of 3
            'TTTCTTTAG', # no start codon
            'ATGTAGCTTTAG', # stop codon in middle
            'ATGTTTTTT', # no stop at end
            'ATGC---TTTAG', # bad insertion
            'ATGCT---TTAG', # bad insertion
            'ATG-CTTTAG', # bad insertion
            'ATG--CTTTAG', # bad insertion
            'ATG----CTTTAG', # bad insertion
        ]

        for seq in bad_seqs:
            padded_sequences['seq2'] = pyfastaq.sequences.Fasta('seq2', seq)
            unpadded_sequences = aln_to_metadata.AlnToMetadata._make_unpadded_seqs(padded_sequences)
            with self.assertRaises(aln_to_metadata.Error):
                aln_to_metadata.AlnToMetadata._check_sequences(padded_sequences, unpadded_sequences, True)


    def test_check_variants_match_sequences(self):
        '''test _check_variants_match_sequences'''
        seqs = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'ATGCTTTAG'),
            'seq2': pyfastaq.sequences.Fasta('seq2', 'ATGCTTCTTTAG'),
            'seq3': pyfastaq.sequences.Fasta('seq3', 'ATG---TAG')
        }

        variants = {'seq1': [(sequence_variant.Variant('p', 'L2M', 'id1'), 'description1')]}
        self.assertTrue(aln_to_metadata.AlnToMetadata._check_variants_match_sequences(seqs, variants, True))
        variants = {'seq1': [(sequence_variant.Variant('p', 'M2L', 'id1'), 'description1')]}
        self.assertTrue(aln_to_metadata.AlnToMetadata._check_variants_match_sequences(seqs, variants, True))

        variants = {'seq1': [(sequence_variant.Variant('p', 'A2M', 'id1'), 'description1')]}
        with self.assertRaises(aln_to_metadata.Error):
            self.assertTrue(aln_to_metadata.AlnToMetadata._check_variants_match_sequences(seqs, variants, True))

        variants = {'seq4': [(sequence_variant.Variant('p', 'A2M', 'id1'), 'description1')]}
        with self.assertRaises(aln_to_metadata.Error):
            self.assertTrue(aln_to_metadata.AlnToMetadata._check_variants_match_sequences(seqs, variants, True))


    def test_variant_ids_are_unique(self):
        '''test variant_ids_are_unique'''
        variants = {
            'seq1': [(sequence_variant.Variant('p', 'L2M', 'id1'), 'description1')],
            'seq2': [(sequence_variant.Variant('p', 'L2M', 'id2'), 'description2')]
        }

        self.assertTrue(aln_to_metadata.AlnToMetadata._variant_ids_are_unique(variants))
        variants['seq2'].append((sequence_variant.Variant('p', 'I3K', 'id1'), 'description3'))
        with self.assertRaises(aln_to_metadata.Error):
            self.assertTrue(aln_to_metadata.AlnToMetadata._variant_ids_are_unique(variants))


    def test_unpadded_to_padded_nt_position(self):
        '''test _unpadded_to_padded_nt_position'''
        ivl = pyfastaq.intervals.Interval

        tests = [
            (0, [], 0),
            (1, [], 1),
            (2, [], 2),
            (0, [ivl(3, 5)], 0),
            (1, [ivl(3, 5)], 1),
            (2, [ivl(3, 5)], 2),
            (3, [ivl(3, 5)], 6),
            (4, [ivl(3, 5)], 7),
            (5, [ivl(3, 5)], 8),
            (0, [ivl(3, 5), ivl(9,14)], 0),
            (1, [ivl(3, 5), ivl(9,14)], 1),
            (2, [ivl(3, 5), ivl(9,14)], 2),
            (3, [ivl(3, 5), ivl(9,14)], 6),
            (4, [ivl(3, 5), ivl(9,14)], 7),
            (5, [ivl(3, 5), ivl(9,14)], 8),
            (6, [ivl(3, 5), ivl(9,14)], 15),
            (7, [ivl(3, 5), ivl(9,14)], 16),
            (8, [ivl(3, 5), ivl(9,14)], 17),
        ]

        for position, insertions, expected in tests:
            got = aln_to_metadata.AlnToMetadata._unpadded_to_padded_nt_position(position, insertions)
            self.assertEqual(expected, got)


    def test_padded_to_unpadded_nt_position(self):
        '''test _padded_to_unpadded_nt_position'''
        ivl = pyfastaq.intervals.Interval

        tests = [
            (0, [], 0),
            (1, [], 1),
            (2, [], 2),
            (0, [ivl(3, 5)], 0),
            (1, [ivl(3, 5)], 1),
            (2, [ivl(3, 5)], 2),
            (3, [ivl(3, 5)], None),
            (4, [ivl(3, 5)], None),
            (5, [ivl(3, 5)], None),
            (6, [ivl(3, 5)], 3),
            (7, [ivl(3, 5)], 4),
            (8, [ivl(3, 5)], 5),
            (0, [ivl(3, 5), ivl(7,10)], 0),
            (1, [ivl(3, 5), ivl(7,10)], 1),
            (2, [ivl(3, 5), ivl(7,10)], 2),
            (3, [ivl(3, 5), ivl(7,10)], None),
            (4, [ivl(3, 5), ivl(7,10)], None),
            (5, [ivl(3, 5), ivl(7,10)], None),
            (6, [ivl(3, 5), ivl(7,10)], 3),
            (7, [ivl(3, 5), ivl(7,10)], None),
            (8, [ivl(3, 5), ivl(7,10)], None),
            (9, [ivl(3, 5), ivl(7,10)], None),
            (10, [ivl(3, 5), ivl(7,10)], None),
            (11, [ivl(3, 5), ivl(7,10)], 4),
            (12, [ivl(3, 5), ivl(7,10)], 5),
        ]

        for position, insertions, expected in tests:
            got = aln_to_metadata.AlnToMetadata._padded_to_unpadded_nt_position(position, insertions)
            self.assertEqual(expected, got)


    def test_variants_to_tsv_lines_coding(self):
        '''test _variants_to_tsv_lines coding sequences'''
        padded_seqs = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'ATG---GCTAATTAG'), # M-AN*
            'seq2': pyfastaq.sequences.Fasta('seq2', 'ATG---GCTAATTAG'), # MFAN*
            'seq3': pyfastaq.sequences.Fasta('seq3', 'ATGTTT---AATTAG'), # MF-N*
            'seq4': pyfastaq.sequences.Fasta('seq4', 'ATGTTTTGTAATTAG'), # MFCN*
            'seq5': pyfastaq.sequences.Fasta('seq5', 'ATGTTTGATAATTAG'), # MFDN*
        }

        unpadded_seqs = aln_to_metadata.AlnToMetadata._make_unpadded_seqs(padded_seqs)
        insertions = aln_to_metadata.AlnToMetadata._make_unpadded_insertion_coords(padded_seqs)

        variant1 = sequence_variant.Variant('p', 'A2D', 'id1')
        variant2 = sequence_variant.Variant('p', 'F2E', 'id2')
        variants = {
            'seq1': [(variant1, 'description 1')],
            'seq5': [(variant2, 'description 2')],
        }

        expected = [
            'seq1\t1\t0\tA2D\tid1\tdescription 1',
            'seq2\t1\t0\tA2D\tid1\tdescription 1',
            'seq4\t1\t0\tC3D\tid1\tdescription 1',
            'seq5\t1\t0\tA3D\tid1\tdescription 1',
            'seq5\t1\t0\tF2E\tid2\tdescription 2',
            'seq3\t1\t0\tF2E\tid2\tdescription 2',
            'seq4\t1\t0\tF2E\tid2\tdescription 2',
        ]

        got = aln_to_metadata.AlnToMetadata._variants_to_tsv_lines(variants, unpadded_seqs, padded_seqs, insertions, True, False)
        self.assertEqual(expected, got)


    def test_variants_to_tsv_lines_noncoding(self):
        '''test _variants_to_tsv_lines noncoding sequences'''
        padded_seqs = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'ATG---GCTAATTAG'),
            'seq2': pyfastaq.sequences.Fasta('seq2', 'ATG---GCTAATTAG'),
            'seq3': pyfastaq.sequences.Fasta('seq3', 'ATGTAT---AATTAG'),
            'seq4': pyfastaq.sequences.Fasta('seq4', 'ATGTGTTGTAATTAG'),
            'seq5': pyfastaq.sequences.Fasta('seq5', 'ATGTTTGATAATTAG'),
        }

        unpadded_seqs = aln_to_metadata.AlnToMetadata._make_unpadded_seqs(padded_seqs)
        insertions = aln_to_metadata.AlnToMetadata._make_unpadded_insertion_coords(padded_seqs)

        variant1 = sequence_variant.Variant('n', 'C5T', 'id1')
        variant2 = sequence_variant.Variant('n', 'A5T', 'id2')
        variants = {
            'seq1': [(variant1, 'description 1')],
            'seq5': [(variant2, 'description 2')],
        }

        expected = [
            'seq1\t0\t1\tC5T\tid1\tdescription 1',
            'seq2\t0\t1\tC5T\tid1\tdescription 1',
            'seq4\t0\t1\tG8T\tid1\tdescription 1',
            'seq5\t0\t1\tA8T\tid1\tdescription 1',
            'seq5\t0\t1\tA5T\tid2\tdescription 2',
            'seq3\t0\t1\tA5T\tid2\tdescription 2',
            'seq4\t0\t1\tG5T\tid2\tdescription 2',
        ]

        got = aln_to_metadata.AlnToMetadata._variants_to_tsv_lines(variants, unpadded_seqs, padded_seqs, insertions, False, True)
        self.assertEqual(expected, got)


    def test_make_cluster_file(self):
        '''test _make_cluster_file'''
        seqs = {
            'seq1': pyfastaq.sequences.Fasta('seq1', 'a'),
            'seq2': pyfastaq.sequences.Fasta('seq2', 'c'),
            'seq3': pyfastaq.sequences.Fasta('seq3', 'g'),
        }
        tmpfile = 'tmp.aln_to_meta_test_make_cluster_file.out'
        expected_file = os.path.join(data_dir, 'aln_to_metadata_make_cluster_file.out')
        aln_to_metadata.AlnToMetadata._make_cluster_file(seqs, tmpfile)
        self.assertTrue(filecmp.cmp(expected_file, tmpfile, shallow=False))
        os.unlink(tmpfile)


    def test_run_coding(self):
        '''test run coding sequences'''
        fa_in = os.path.join(data_dir, 'aln_to_metadata_run_coding.in.fa')
        fa_expected = os.path.join(data_dir, 'aln_to_metadata_run_coding.out.fa')
        tsv_in = os.path.join(data_dir, 'aln_to_metadata_run_coding.in.tsv')
        tsv_expected = os.path.join(data_dir, 'aln_to_metadata_run_coding.out.tsv')
        cluster_expected = os.path.join(data_dir, 'aln_to_metadata_run_coding.out.cluster')
        a_to_m = aln_to_metadata.AlnToMetadata(fa_in, tsv_in, True, False)
        outprefix = 'tmp.test.aln_to_metadata.run_coding'
        a_to_m.run(outprefix)
        self.assertTrue(filecmp.cmp(tsv_expected, outprefix + '.tsv', shallow=False))
        self.assertTrue(filecmp.cmp(fa_expected, outprefix + '.fa', shallow=False))
        self.assertTrue(filecmp.cmp(cluster_expected, outprefix + '.cluster', shallow=False))
        os.unlink(outprefix + '.tsv')
        os.unlink(outprefix + '.fa')
        os.unlink(outprefix + '.cluster')


    def test_run_noncoding(self):
        '''test run noncoding sequences'''
        fa_in = os.path.join(data_dir, 'aln_to_metadata_run_noncoding.in.fa')
        fa_expected = os.path.join(data_dir, 'aln_to_metadata_run_noncoding.out.fa')
        tsv_in = os.path.join(data_dir, 'aln_to_metadata_run_noncoding.in.tsv')
        tsv_expected = os.path.join(data_dir, 'aln_to_metadata_run_noncoding.out.tsv')
        cluster_expected = os.path.join(data_dir, 'aln_to_metadata_run_noncoding.out.cluster')
        a_to_m = aln_to_metadata.AlnToMetadata(fa_in, tsv_in, False, True)
        outprefix = 'tmp.test.aln_to_metadata.run_noncoding'
        a_to_m.run(outprefix)
        self.assertTrue(filecmp.cmp(tsv_expected, outprefix + '.tsv', shallow=False))
        self.assertTrue(filecmp.cmp(fa_expected, outprefix + '.fa', shallow=False))
        self.assertTrue(filecmp.cmp(cluster_expected, outprefix + '.cluster', shallow=False))
        os.unlink(outprefix + '.tsv')
        os.unlink(outprefix + '.fa')
        os.unlink(outprefix + '.cluster')




import unittest
import os
import shutil
import filecmp
import pyfastaq
from ariba import read_store

modules_dir = os.path.dirname(os.path.abspath(read_store.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')


def file_to_list(infile):
    f = pyfastaq.utils.open_file_read(infile)
    lines = [x for x in f.readlines()]
    pyfastaq.utils.close(f)
    return lines


class TestReadStore(unittest.TestCase):
    def test_sort_file(self):
        '''test _sort_file'''
        infile = os.path.join(data_dir, 'read_store_test_sort_file.in')
        expected = os.path.join(data_dir, 'read_store_test_sort_file.out')
        tmpfile = 'tmp.read_store_test_sort_file.out'
        read_store.ReadStore._sort_file(infile, tmpfile)
        self.assertTrue(filecmp.cmp(expected, tmpfile, shallow=False))
        os.unlink(tmpfile)


    def test_compress_and_index_file(self):
        '''Test _compress_and_index_file'''
        infile = os.path.join(data_dir, 'read_store_test_compress_and_index_file.in')
        tmpfile = 'tmp.test_compress_and_index_file.in'
        tmpfile_gz = 'tmp.test_compress_and_index_file.in.gz'
        shutil.copyfile(infile, tmpfile)
        read_store.ReadStore._compress_and_index_file(tmpfile)
        self.assertTrue(os.path.exists(tmpfile_gz))
        expected_lines = file_to_list(infile)
        got_lines = file_to_list(tmpfile_gz)
        self.assertEqual(expected_lines, got_lines)
        self.assertTrue(os.path.exists(tmpfile_gz + '.tbi'))
        os.unlink(tmpfile)
        os.unlink(tmpfile_gz)
        os.unlink(tmpfile_gz + '.tbi')


    def test_get_reads_fq_pair(self):
        '''Test get_reads fastq pair'''
        infile = os.path.join(data_dir, 'read_store_test_get_reads.in')
        expected1 = os.path.join(data_dir, 'read_store_test_get_reads.expected.reads_1.fq')
        expected2 = os.path.join(data_dir, 'read_store_test_get_reads.expected.reads_2.fq')
        outprefix = 'tmp.read_store_test_get_reads'
        reads1 = outprefix + '.reads_1.fq'
        reads2 = outprefix + '.reads_2.fq'
        rstore = read_store.ReadStore(infile, outprefix)
        got_reads, got_bases = rstore.get_reads('cluster2', reads1, out2=reads2)
        self.assertEqual(6, got_reads)
        self.assertEqual(24, got_bases)
        self.assertTrue(filecmp.cmp(expected1, reads1))
        self.assertTrue(filecmp.cmp(expected2, reads2))
        os.unlink(outprefix + '.gz')
        os.unlink(outprefix + '.gz.tbi')
        os.unlink(reads1)
        os.unlink(reads2)


    def test_get_reads_fq_interleave(self):
        '''Test get_reads fastq interleaved'''
        infile = os.path.join(data_dir, 'read_store_test_get_reads.in')
        expected = os.path.join(data_dir, 'read_store_test_get_reads.expected.reads.fq')
        outprefix = 'tmp.read_store_test_get_reads'
        reads = outprefix + '.reads_1.fq'
        rstore = read_store.ReadStore(infile, outprefix)
        got_reads, got_bases = rstore.get_reads('cluster2', reads)
        self.assertEqual(6, got_reads)
        self.assertEqual(24, got_bases)
        self.assertTrue(filecmp.cmp(expected, reads))
        os.unlink(outprefix + '.gz')
        os.unlink(outprefix + '.gz.tbi')
        os.unlink(reads)


    def test_get_reads_fa_pair(self):
        '''Test get_reads fasta pair'''
        infile = os.path.join(data_dir, 'read_store_test_get_reads.in')
        expected1 = os.path.join(data_dir, 'read_store_test_get_reads.expected.reads_1.fa')
        expected2 = os.path.join(data_dir, 'read_store_test_get_reads.expected.reads_2.fa')
        outprefix = 'tmp.read_store_test_get_reads'
        reads1 = outprefix + '.reads_1.fa'
        reads2 = outprefix + '.reads_2.fa'
        rstore = read_store.ReadStore(infile, outprefix)
        got_reads, got_bases = rstore.get_reads('cluster2', reads1, out2=reads2, fasta=True)
        self.assertEqual(6, got_reads)
        self.assertEqual(24, got_bases)
        self.assertTrue(filecmp.cmp(expected1, reads1))
        self.assertTrue(filecmp.cmp(expected2, reads2))
        os.unlink(outprefix + '.gz')
        os.unlink(outprefix + '.gz.tbi')
        os.unlink(reads1)
        os.unlink(reads2)


    def test_get_reads_subset(self):
        '''Test get_reads subset'''
        infile = os.path.join(data_dir, 'read_store_test_get_reads.in')
        expected1 = os.path.join(data_dir, 'read_store_test_get_reads.expected.reads_subset.1.fq')
        expected2 = os.path.join(data_dir, 'read_store_test_get_reads.expected.reads_subset.2.fq')
        wanted_ids = {1, 11}
        outprefix = 'tmp.read_store_test_get_reads'
        reads1 = outprefix + '.reads_1.fq'
        reads2 = outprefix + '.reads_2.fq'
        rstore = read_store.ReadStore(infile, outprefix)
        got_reads, got_bases = rstore.get_reads('cluster2', reads1, out2=reads2, wanted_ids=wanted_ids)
        self.assertEqual(4, got_reads)
        self.assertEqual(16, got_bases)
        self.assertTrue(filecmp.cmp(expected1, reads1))
        self.assertTrue(filecmp.cmp(expected2, reads2))
        os.unlink(outprefix + '.gz')
        os.unlink(outprefix + '.gz.tbi')
        os.unlink(reads1)
        os.unlink(reads2)


    def test_clean(self):
        '''Test clean'''
        infile = os.path.join(data_dir, 'read_store_test_clean.in')
        outprefix = 'tmp.read_store_test_clean'
        self.assertFalse(os.path.exists(outprefix))
        self.assertFalse(os.path.exists(outprefix + '.gz'))
        self.assertFalse(os.path.exists(outprefix + '.gz.tbi'))
        rstore = read_store.ReadStore(infile, outprefix)
        self.assertFalse(os.path.exists(outprefix))
        self.assertTrue(os.path.exists(outprefix + '.gz'))
        self.assertTrue(os.path.exists(outprefix + '.gz.tbi'))
        rstore.clean()
        self.assertFalse(os.path.exists(outprefix))
        self.assertFalse(os.path.exists(outprefix + '.gz'))
        self.assertFalse(os.path.exists(outprefix + '.gz.tbi'))



import os
import sys
import csv
import zipfile
import pyfastaq
from ariba import common

class Error (Exception): pass


class MegaresZipParser:
    def __init__(self, zip_url, outprefix):
        self.zip_url = zip_url
        self.outprefix = outprefix
        self.zip_file = self.outprefix + '.downloaded.zip'


    @classmethod
    def _extract_files(cls, zip_file, outdir):
        original_files = {'annotations': None, 'fasta': None, 'header_mappings': None}

        try:
            os.mkdir(outdir)
        except:
            raise Error('Error making directory ' + outdir)

        # Old <2.0.0 megares has eg these files:
        #  megares_annotations_v1.01.csv
        #  megares_database_v1.01.fasta
        #  megares_to_external_header_mappings_v1.01.tsv
        # megares 2.0.0 has these files:
        #  megares_drugs_annotations_v2.00.csv
        #  megares_drugs_database_v2.00.fasta
        #  megares_modified_annotations_v2.00.csv
        #  megares_modified_database_v2.00.fasta
        #  megares_to_external_header_mappings_v2.00.csv
        # The sequences in *_modified_* files seem to be a superset of
        # *_drugs_*, so use the *_modified_* ones. This will happen
        # as long as we loop over sorted filenames, because the _modified_
        # csv and fasta are listed last
        zfile = zipfile.ZipFile(zip_file)
        for member in sorted(zfile.namelist()):
            if '_annotations_' in member:
                original_files['annotations'] = member
            elif '_database_' in member and member.endswith('.fasta'):
                original_files['fasta'] = member
            elif '_header_mappings_' in member:
                original_files['header_mappings'] = member
            else:
                continue

            zfile.extract(member, path=outdir)

        if None in original_files.values():
            common.rmtree(outdir)
            raise Error('Error. Not all expected files found in downloaded megares zipfile. ' + str(original_files))

        return original_files


    @classmethod
    def _csv_to_dict(cls, infile, delimiter, expected_columns, key_column):
        data = {}
        non_key_columns = expected_columns - {key_column}

        with open(infile) as f:
            reader = csv.DictReader(f, delimiter=delimiter)
            if not set(expected_columns).issubset(set(reader.fieldnames)):
                raise Error('Unexpected header in annotations file. Expected columns: ' + ','.join(expected_columns) + ' but got: ' + ','.join(reader.fieldnames))

            for row in reader:
                data[row[key_column]] = {x: row[x] for x in non_key_columns}

        return data


    @classmethod
    def _load_annotations_file(cls, infile):
        return MegaresZipParser._csv_to_dict(infile, ',', {'header', 'class', 'mechanism', 'group'}, 'header')


    @classmethod
    def _load_header_mappings_file(cls, infile):
        # Megares <2.0.0 uses a tsv file, whereas 2.0.0 uses csv.
        # Also, the column names changed slightly for 2.0.0, so we'll change
        # them to be the same as <2.0.0 after loading the file
        if infile.endswith(".tsv"):
            return MegaresZipParser._csv_to_dict(infile, '\t', {'Source_Database', 'MEGARes_Header', 'Source_Headers(space_separated)'}, 'MEGARes_Header')
        else:
            assert infile.endswith(".csv")
            data = MegaresZipParser._csv_to_dict(infile, ',', {'Database', 'MEGARes_v2_header', 'Source_header'}, 'MEGARes_v2_header')
            fixed_data = {}
            for key, d in data.items():
                fixed_data[key] = {
                    "Source_Database": d["Database"],
                    "Source_Headers(space_separated)": d["Source_header"]
                }
            return fixed_data


    @classmethod
    def _write_files(cls, outprefix, sequences, annotations, header_mappings):
        fasta = outprefix + '.fa'
        tsv = outprefix + '.tsv'
        fh_fasta = pyfastaq.utils.open_file_write(fasta)
        fh_tsv = pyfastaq.utils.open_file_write(tsv)

        for seq in sorted(sequences):
            final_column = []

            if seq in annotations:
                group = annotations[seq]['group']
                final_column.append('class:' + annotations[seq]['class'] + '; mechanism:' + annotations[seq]['mechanism'] + '; group:' + group)
            else:
                group = 'unknown'
                print('WARNING: sequence "', seq, '" has no record in annotations file', sep='', file=sys.stderr)

            if seq in header_mappings:
                final_column.append('Source_Database:' + header_mappings[seq]['Source_Database'] + '; Source_Headers:' + header_mappings[seq]['Source_Headers(space_separated)'])
            else:
                print('WARNING: sequence "', seq, '" has no record in header mappings file', sep='', file=sys.stderr)

            if len(final_column) > 0:
                print(group + '.' + seq, '1', '0', '.', '.', '; '.join(final_column), sep='\t', file=fh_tsv)
            else:
                print(group + '.' + seq, '1', '0', '.', '.', '.', sep='\t', file=fh_tsv)

            sequences[seq].id = group + '.' + sequences[seq].id
            print(sequences[seq], file=fh_fasta)

        fh_fasta.close()
        fh_tsv.close()


    def run(self):
        common.download_file(self.zip_url, self.zip_file, verbose=True)
        tmpdir = self.zip_file + '.tmp.extract'
        original_files = MegaresZipParser._extract_files(self.zip_file, tmpdir)
        annotation_data = MegaresZipParser._load_annotations_file(os.path.join(tmpdir, original_files['annotations']))
        header_data = MegaresZipParser._load_header_mappings_file(os.path.join(tmpdir, original_files['header_mappings']))
        sequences = {}
        pyfastaq.tasks.file_to_dict(os.path.join(tmpdir, original_files['fasta']), sequences)
        MegaresZipParser._write_files(self.outprefix, sequences, annotation_data, header_data)
        common.rmtree(tmpdir)
        os.unlink(self.zip_file)




import pyfastaq
from ariba import report, flag

class Error (Exception): pass

class ReportFilter:
    def __init__(self,
            infile=None,
            min_pc_ident=90,
            min_ref_base_assembled=1,
            ignore_not_has_known_variant=False,
            remove_synonymous_snps=True,
            exclude_flags=None,
        ):

        if infile is not None:
            self.report = self._load_report(infile)
        else:
            self.report = {}

        self.min_pc_ident = min_pc_ident
        self.min_ref_base_assembled = min_ref_base_assembled
        self.ignore_not_has_known_variant = ignore_not_has_known_variant
        self.remove_synonymous_snps = remove_synonymous_snps

        if exclude_flags is None:
            self.exclude_flags = ['assembly_fail', 'ref_seq_choose_fail']
        else:
            self.exclude_flags = exclude_flags


    @classmethod
    def _report_line_to_dict(cls, line):
        '''Takes report line string as input. Returns a dict of column name -> value in line'''
        data = line.split('\t')
        if len(data) != len(report.columns):
            return None

        d = dict(zip(report.columns, data))
        for key in report.int_columns:
            try:
                d[key] = int(d[key])
            except:
                assert d[key] == '.'

        for key in report.float_columns:
            try:
                d[key] = float(d[key])
            except:
                assert d[key] == '.'

        d['flag'] = flag.Flag(int(d['flag']))
        return d


    @classmethod
    def _dict_to_report_line(cls, report_dict):
        '''Takes a report_dict as input and returns a report line'''
        return '\t'.join([str(report_dict[x]) for x in report.columns])


    @staticmethod
    def _load_report(infile):
        '''Loads report file into a dictionary. Key=reference name.
        Value = list of report lines for that reference'''
        report_dict = {}
        f = pyfastaq.utils.open_file_read(infile)
        first_line = True

        for line in f:
            line = line.rstrip()

            if first_line:
                expected_first_line = '#' + '\t'.join(report.columns)
                if line != expected_first_line:
                    pyfastaq.utils.close(f)
                    raise Error('Error reading report file. Expected first line of file is\n' + expected_first_line + '\nbut got:\n' + line)
                first_line = False
            else:
                line_dict = ReportFilter._report_line_to_dict(line)
                if line_dict is None:
                    pyfastaq.utils.close(f)
                    raise Error('Error reading report file at this line:\n' + line)
                ref_name = line_dict['ref_name']
                ctg_name = line_dict['ctg']
                if ref_name not in report_dict:
                    report_dict[ref_name] = {}
                if ctg_name not in report_dict[ref_name]:
                    report_dict[ref_name][ctg_name] = []

                report_dict[ref_name][ctg_name].append(line_dict)

        pyfastaq.utils.close(f)
        return report_dict


    @staticmethod
    def _flag_passes_filter(flag, exclude_flags):
        for f in exclude_flags:
            if flag.has(f):
                return False
        return True


    def _report_dict_passes_non_essential_filters(self, report_dict):
        # known_var == '.' iff this line is not reporting a variant. Which means it passes all non-essential filters
        if report_dict.get('known_var', '.') == '.':
            return True

        if self.remove_synonymous_snps and report_dict.get('ref_ctg_effect', None) == 'SYN':
            return False

        if self.ignore_not_has_known_variant and report_dict['known_var'] == '1' and report_dict['has_known_var'] == '0':
            return False

        return True


    def _report_dict_passes_essential_filters(self, report_dict):
        return ReportFilter._flag_passes_filter(report_dict['flag'], self.exclude_flags) \
                   and report_dict['pc_ident'] >= self.min_pc_ident \
                   and report_dict['ref_base_assembled'] >= self.min_ref_base_assembled


    def _filter_list_of_dicts(self, dicts_list):
        if len(dicts_list) == 0:
            return []

        pass_dicts = []
        essential_dicts = []
        fail_dicts = []

        for d in dicts_list:
            if self._report_dict_passes_essential_filters(d):
                if self._report_dict_passes_non_essential_filters(d):
                    pass_dicts.append(d)
                else:
                    essential_dicts.append(d)
            else:
                fail_dicts.append(d)

        if len(pass_dicts) == 0:
            assert len(fail_dicts) + len(essential_dicts) > 0
            if len(essential_dicts) > 0:
                new_d = essential_dicts[0]
                for key in report.var_columns:
                    new_d[key] = '.'
                pass_dicts.append(new_d)

        return ReportFilter._remove_all_after_first_frameshift(pass_dicts)


    @staticmethod
    def _remove_all_after_first_frameshift(dicts_list):
        if len(dicts_list) == 0 or dicts_list[0]['flag'].has('complete_gene'):
            return dicts_list

        fshift_starts = [int(d['ref_start']) for d in dicts_list if d.get('ref_ctg_effect', None) == 'FSHIFT']
        if len(fshift_starts) == 0:
            return dicts_list

        first_start = min(fshift_starts)

        return [d for d in dicts_list if d['ref_start'] == '.' or d['ref_start'] <= first_start]


    def _filter_dicts(self):
        '''Filters out all the report_dicts that do not pass the cutoffs. If any ref sequence
           loses all of its report_dicts, then it is completely removed.'''
        keys_to_remove = set()

        for ref_name in self.report:
            for ctg_name in self.report[ref_name]:
                self.report[ref_name][ctg_name] = self._filter_list_of_dicts(self.report[ref_name][ctg_name])
                if len(self.report[ref_name][ctg_name]) == 0:
                    keys_to_remove.add((ref_name, ctg_name))

        refs_to_remove = set()

        for ref_name, ctg_name in keys_to_remove:
            del self.report[ref_name][ctg_name]
            if len(self.report[ref_name]) == 0:
                refs_to_remove.add(ref_name)

        for ref_name in refs_to_remove:
            del self.report[ref_name]


    def _write_report_tsv(self, outfile):
        f = pyfastaq.utils.open_file_write(outfile)
        print('#' + '\t'.join(report.columns), file=f)

        for ref_name in sorted(self.report):
            for ctg_name, report_dicts in sorted(self.report[ref_name].items()):
                for d in report_dicts:
                    print(ReportFilter._dict_to_report_line(d), file=f)

        pyfastaq.utils.close(f)


    def run(self, outfile):
        self._filter_dicts()
        self._write_report_tsv(outfile)

